@@ -21,9 +21,10 @@
 
 #include "private-libwebsockets.h"
 
-int lws_handshake_client(struct lws *wsi, unsigned char **buf, size_t len)
+int
+lws_handshake_client(struct lws *wsi, unsigned char **buf, size_t len)
 {
-	unsigned int n;
+	int m;
 
 	switch (wsi->mode) {
 	case LWSCM_WSCL_WAITING_PROXY_REPLY:
@@ -31,22 +32,47 @@ int lws_handshake_client(struct lws *wsi, unsigned char **buf, size_t len)
 	case LWSCM_WSCL_WAITING_SERVER_REPLY:
 	case LWSCM_WSCL_WAITING_EXTENSION_CONNECT:
 	case LWSCM_WS_CLIENT:
-		for (n = 0; n < len; n++)
-			if (lws_client_rx_sm(wsi, *(*buf)++)) {
-				lwsl_debug("client_rx_sm failed\n");
-				return 1;
+		while (len) {
+			/*
+			 * we were accepting input but now we stopped doing so
+			 */
+			if (!(wsi->rxflow_change_to & LWS_RXFLOW_ALLOW)) {
+				lwsl_debug("%s: caching %d\n", __func__, len);
+				lws_rxflow_cache(wsi, *buf, 0, len);
+				return 0;
 			}
+			if (wsi->u.ws.rx_draining_ext) {
+				m = lws_rx_sm(wsi, 0);
+				if (m < 0)
+					return -1;
+				continue;
+			}
+			/* account for what we're using in rxflow buffer */
+			if (wsi->rxflow_buffer)
+				wsi->rxflow_pos++;
+
+			if (lws_client_rx_sm(wsi, *(*buf)++)) {
+				lwsl_debug("client_rx_sm exited\n");
+				return -1;
+			}
+			len--;
+		}
+		lwsl_debug("%s: finished with %d\n", __func__, len);
 		return 0;
 	default:
 		break;
 	}
+
 	return 0;
 }
 
-int lws_client_socket_service(struct lws_context *context,
-			      struct lws *wsi, struct lws_pollfd *pollfd)
+int
+lws_client_socket_service(struct lws_context *context, struct lws *wsi,
+			  struct lws_pollfd *pollfd)
 {
-	char *p = (char *)&context->serv_buf[0];
+	struct lws_context_per_thread *pt = &context->pt[(int)wsi->tsi];
+	char *p = (char *)&pt->serv_buf[0];
+	char *sb = p;
 	unsigned char c;
 	int n, len;
 
@@ -59,7 +85,7 @@ int lws_client_socket_service(struct lws_context *context,
 		 * timeout protection set in client-handshake.c
 		 */
 
-               if (lws_client_connect_2(wsi) == NULL) {
+		if (!lws_client_connect_2(wsi)) {
 			/* closed */
 			lwsl_client("closed\n");
 			return -1;
@@ -81,10 +107,8 @@ int lws_client_socket_service(struct lws_context *context,
 			return 0;
 		}
 
-		n = recv(wsi->sock, (char *)context->serv_buf,
-					sizeof(context->serv_buf), 0);
+		n = recv(wsi->sock, sb, LWS_MAX_SOCKET_IO_BUF, 0);
 		if (n < 0) {
-
 			if (LWS_ERRNO == LWS_EAGAIN) {
 				lwsl_debug("Proxy read returned EAGAIN... retrying\n");
 				return 0;
@@ -95,12 +119,11 @@ int lws_client_socket_service(struct lws_context *context,
 			return 0;
 		}
 
-		context->serv_buf[13] = '\0';
-		if (strcmp((char *)context->serv_buf, "HTTP/1.0 200 ") &&
-		    strcmp((char *)context->serv_buf, "HTTP/1.1 200 ")
-		) {
+		pt->serv_buf[13] = '\0';
+		if (strcmp(sb, "HTTP/1.0 200 ") &&
+		    strcmp(sb, "HTTP/1.1 200 ")) {
 			lws_close_free_wsi(wsi, LWS_CLOSE_STATUS_NOSTATUS);
-			lwsl_err("ERROR proxy: %s\n", context->serv_buf);
+			lwsl_err("ERROR proxy: %s\n", sb);
 			return 0;
 		}
 
@@ -115,15 +138,12 @@ int lws_client_socket_service(struct lws_context *context,
 		/*
 		 * we are under PENDING_TIMEOUT_SENT_CLIENT_HANDSHAKE
 		 * timeout protection set in client-handshake.c
-		 */
-
-		/*
+		 *
 		 * take care of our lws_callback_on_writable
 		 * happening at a time when there's no real connection yet
 		 */
 		if (lws_change_pollfd(wsi, LWS_POLLOUT, 0))
 			return -1;
-		lws_libev_io(wsi, LWS_EV_STOP | LWS_EV_WRITE);
 
 #ifdef LWS_OPENSSL_SUPPORT
 		/* we can retry this... just cook the SSL BIO the first time */
@@ -204,8 +224,7 @@ int lws_client_socket_service(struct lws_context *context,
 			lws_latency_pre(context, wsi);
 			n = SSL_connect(wsi->ssl);
 			lws_latency(context, wsi,
-			  "SSL_connect LWSCM_WSCL_ISSUE_HANDSHAKE",
-								      n, n > 0);
+			  "SSL_connect LWSCM_WSCL_ISSUE_HANDSHAKE", n, n > 0);
 
 			if (n < 0) {
 				n = SSL_get_error(wsi->ssl, n);
@@ -226,9 +245,7 @@ int lws_client_socket_service(struct lws_context *context,
 					 * are getting serviced inbetweentimes)
 					 * us to get called back when writable.
 					 */
-
-					lwsl_info(
-					     "SSL_connect WANT_WRITE... retrying\n");
+					lwsl_info("%s: WANT_WRITE... retrying\n", __func__);
 					lws_callback_on_writable(wsi);
 some_wait:
 					wsi->mode = LWSCM_WSCL_WAITING_SSL;
@@ -243,13 +260,10 @@ some_wait:
 				 * retry if new data comes until we
 				 * run into the connection timeout or win
 				 */
-
 				n = ERR_get_error();
 				if (n != SSL_ERROR_NONE) {
 					lwsl_err("SSL connect error %lu: %s\n",
-						n,
-						ERR_error_string(n,
-							  (char *)context->serv_buf));
+						n, ERR_error_string(n, sb));
 					return 0;
 				}
 			}
@@ -261,7 +275,6 @@ some_wait:
 	case LWSCM_WSCL_WAITING_SSL:
 
 		if (wsi->use_ssl) {
-
 			if (wsi->mode == LWSCM_WSCL_WAITING_SSL) {
 				lws_latency_pre(context, wsi);
 				n = SSL_connect(wsi->ssl);
@@ -288,7 +301,6 @@ some_wait:
 						 * are getting serviced inbetweentimes)
 						 * us to get called back when writable.
 						 */
-
 						lwsl_info("SSL_connect WANT_WRITE... retrying\n");
 						lws_callback_on_writable(wsi);
 
@@ -305,8 +317,7 @@ some_wait:
 					n = ERR_get_error();
 					if (n != SSL_ERROR_NONE) {
 						lwsl_err("SSL connect error %lu: %s\n",
-							 n, ERR_error_string(n,
-							 (char *)context->serv_buf));
+							 n, ERR_error_string(n, sb));
 						return 0;
 					}
 				}
@@ -329,7 +340,7 @@ some_wait:
 					lwsl_notice("accepting self-signed certificate\n");
 				} else {
 					lwsl_err("server's cert didn't look good, X509_V_ERR = %d: %s\n",
-						 n, ERR_error_string(n, (char *)context->serv_buf));
+						 n, ERR_error_string(n, sb));
 					lws_close_free_wsi(wsi,
 						LWS_CLOSE_STATUS_NOSTATUS);
 					return 0;
@@ -342,7 +353,7 @@ some_wait:
 
 		wsi->mode = LWSCM_WSCL_ISSUE_HANDSHAKE2;
 		lws_set_timeout(wsi, PENDING_TIMEOUT_AWAITING_CLIENT_HS_SEND,
-				AWAITING_TIMEOUT);
+				context->timeout_secs);
 
 		/* fallthru */
 
@@ -358,10 +369,9 @@ some_wait:
 
 		lws_latency_pre(context, wsi);
 
-		n = lws_ssl_capable_write(wsi, context->serv_buf,
-					  p - (char *)context->serv_buf);
+		n = lws_ssl_capable_write(wsi, (unsigned char *)sb, p - sb);
 		lws_latency(context, wsi, "send lws_issue_raw", n,
-			    n == p - (char *)context->serv_buf);
+			    n == p - sb);
 		switch (n) {
 		case LWS_SSL_CAPABLE_ERROR:
 			lwsl_debug("ERROR writing to client socket\n");
@@ -376,7 +386,7 @@ some_wait:
 		wsi->u.hdr.lextable_pos = 0;
 		wsi->mode = LWSCM_WSCL_WAITING_SERVER_REPLY;
 		lws_set_timeout(wsi, PENDING_TIMEOUT_AWAITING_SERVER_RESPONSE,
-				AWAITING_TIMEOUT);
+				context->timeout_secs);
 		break;
 
 	case LWSCM_WSCL_WAITING_SERVER_REPLY:
@@ -420,6 +430,7 @@ some_wait:
 			lws_latency(context, wsi, "send lws_issue_raw", n,
 				    n == 1);
 			switch (n) {
+			case 0:
 			case LWS_SSL_CAPABLE_ERROR:
 				goto bail3;
 			case LWS_SSL_CAPABLE_MORE_SERVICE:
@@ -468,7 +479,6 @@ bail3:
 	return 0;
 }
 
-
 /*
  * In-place str to lower case
  */
@@ -485,15 +495,19 @@ strtolower(char *s)
 int
 lws_client_interpret_server_handshake(struct lws *wsi)
 {
+	int n, len, okay = 0, isErrorCodeReceived = 0, port = 0, ssl = 0;
 	struct lws_context *context = wsi->context;
 	int close_reason = LWS_CLOSE_STATUS_PROTOCOL_ERR;
-	int n, len, okay = 0, isErrorCodeReceived = 0;
-	const char *pc;
+	const char *pc, *prot, *ads = NULL, *path;
 	char *p;
 #ifndef LWS_NO_EXTENSIONS
+	struct lws_context_per_thread *pt = &context->pt[(int)wsi->tsi];
+	char *sb = (char *)&pt->serv_buf[0];
+	const struct lws_ext_options *opts;
 	const struct lws_extension *ext;
 	char ext_name[128];
-	const char *c;
+	const char *c, *a;
+	char ignore;
 	int more = 1;
 	void *v;
 #endif
@@ -502,13 +516,29 @@ lws_client_interpret_server_handshake(struct lws *wsi)
 	 * well, what the server sent looked reasonable for syntax.
 	 * Now let's confirm it sent all the necessary headers
 	 */
-
 	p = lws_hdr_simple_ptr(wsi, WSI_TOKEN_HTTP);
 	if (!p) {
 		lwsl_info("no URI\n");
 		goto bail3;
 	}
+	n = atoi(p);
+	if (n == 301 || n == 302 || n == 303 || n == 307 || n == 308) {
+		p = lws_hdr_simple_ptr(wsi, WSI_TOKEN_HTTP_LOCATION);
+		if (!p)
+			goto bail3;
 
+		if (lws_parse_uri(p, &prot, &ads, &port, &path))
+			goto bail3;
+
+		if (!strcmp(prot, "wss://") || !strcmp(prot, "https://"))
+			ssl = 1;
+
+		if (lws_client_reset(wsi, ssl, ads, port, path, ads)) {
+			lwsl_err("Redirect failed\n");
+			goto bail3;
+		}
+		return 0;
+	}
 	if (lws_hdr_total_length(wsi, WSI_TOKEN_ACCEPT) == 0) {
 		lwsl_info("no ACCEPT\n");
 		isErrorCodeReceived = 1;
@@ -557,7 +587,6 @@ lws_client_interpret_server_handshake(struct lws *wsi)
 
 	len = lws_hdr_total_length(wsi, WSI_TOKEN_PROTOCOL);
 	if (!len) {
-
 		lwsl_info("lws_client_int_s_hs: WSI_TOKEN_PROTOCOL is null\n");
 		/*
 		 * no protocol name to work from,
@@ -611,7 +640,7 @@ check_extensions:
 	/* instantiate the accepted extensions */
 
 	if (!lws_hdr_total_length(wsi, WSI_TOKEN_EXTENSIONS)) {
-		lwsl_ext("no client extenstions allowed by server\n");
+		lwsl_ext("no client extensions allowed by server\n");
 		goto check_accept;
 	}
 
@@ -620,23 +649,35 @@ check_extensions:
 	 * and go through matching them or identifying bogons
 	 */
 
-	if (lws_hdr_copy(wsi, (char *)context->serv_buf,
-		   sizeof(context->serv_buf), WSI_TOKEN_EXTENSIONS) < 0) {
+	if (lws_hdr_copy(wsi, sb, LWS_MAX_SOCKET_IO_BUF, WSI_TOKEN_EXTENSIONS) < 0) {
 		lwsl_warn("ext list from server failed to copy\n");
 		goto bail2;
 	}
 
-	c = (char *)context->serv_buf;
+	c = sb;
 	n = 0;
+	ignore = 0;
+	a = NULL;
 	while (more) {
 
-		if (*c && (*c != ',' && *c != ' ' && *c != '\t')) {
+		if (*c && (*c != ',' && *c != '\t')) {
+			if (*c == ';') {
+				ignore = 1;
+				if (!a)
+					a = c + 1;
+			}
+			if (ignore || *c == ' ') {
+				c++;
+				continue;
+			}
+
 			ext_name[n] = *c++;
 			if (n < sizeof(ext_name) - 1)
 				n++;
 			continue;
 		}
 		ext_name[n] = '\0';
+		ignore = 0;
 		if (!*c)
 			more = 0;
 		else {
@@ -647,7 +688,7 @@ check_extensions:
 
 		/* check we actually support it */
 
-		lwsl_ext("checking client ext %s\n", ext_name);
+		lwsl_notice("checking client ext %s\n", ext_name);
 
 		n = 0;
 		ext = lws_get_context(wsi)->extensions;
@@ -658,30 +699,60 @@ check_extensions:
 			}
 
 			n = 1;
-			lwsl_ext("instantiating client ext %s\n", ext_name);
+			lwsl_notice("instantiating client ext %s\n", ext_name);
 
 			/* instantiate the extension on this conn */
 
-			wsi->active_extensions_user[
-				wsi->count_active_extensions] =
-					 lws_zalloc(ext->per_session_data_size);
-			if (wsi->active_extensions_user[
-				wsi->count_active_extensions] == NULL) {
-				lwsl_err("Out of mem\n");
-				goto bail2;
-			}
-			wsi->active_extensions[
-				  wsi->count_active_extensions] = ext;
+			wsi->active_extensions[wsi->count_act_ext] = ext;
 
-			/* allow him to construct his context */
+			/* allow him to construct his ext instance */
 
 			ext->callback(lws_get_context(wsi), ext, wsi,
-				      LWS_EXT_CALLBACK_CLIENT_CONSTRUCT,
-				      wsi->active_extensions_user[
-					 wsi->count_active_extensions],
-				      NULL, 0);
+				      LWS_EXT_CB_CLIENT_CONSTRUCT,
+				      (void *)&wsi->act_ext_user[wsi->count_act_ext],
+				      (void *)&opts, 0);
 
-			wsi->count_active_extensions++;
+			/*
+			 * allow the user code to override ext defaults if it
+			 * wants to
+			 */
+			ext_name[0] = '\0';
+			if (user_callback_handle_rxflow(wsi->protocol->callback,
+					wsi, LWS_CALLBACK_WS_EXT_DEFAULTS,
+					(char *)ext->name, ext_name,
+					sizeof(ext_name)))
+				goto bail2;
+
+			if (ext_name[0] &&
+			    lws_ext_parse_options(ext, wsi, wsi->act_ext_user[
+						  wsi->count_act_ext], opts, ext_name,
+						  strlen(ext_name))) {
+				lwsl_err("%s: unable to parse user defaults '%s'",
+					 __func__, ext_name);
+				goto bail2;
+			}
+
+			/*
+			 * give the extension the server options
+			 */
+			if (a && lws_ext_parse_options(ext, wsi,
+					wsi->act_ext_user[wsi->count_act_ext],
+					opts, a, c - a)) {
+				lwsl_err("%s: unable to parse remote def '%s'",
+					 __func__, a);
+				goto bail2;
+			}
+
+			if (ext->callback(lws_get_context(wsi), ext, wsi,
+					LWS_EXT_CB_OPTION_CONFIRM,
+				      wsi->act_ext_user[wsi->count_act_ext],
+				      NULL, 0)) {
+				lwsl_err("%s: ext %s rejects server options %s",
+					 ext->name, a);
+				goto bail2;
+			}
+
+			wsi->count_act_ext++;
 
 			ext++;
 		}
@@ -691,6 +762,7 @@ check_extensions:
 			goto bail2;
 		}
 
+		a = NULL;
 		n = 0;
 	}
 
@@ -718,18 +790,15 @@ check_accept:
 	 * we seem to be good to go, give client last chance to check
 	 * headers and OK it
 	 */
-	if (wsi->protocol->callback(wsi,
-				    LWS_CALLBACK_CLIENT_FILTER_PRE_ESTABLISH,
+	if (wsi->protocol->callback(wsi, LWS_CALLBACK_CLIENT_FILTER_PRE_ESTABLISH,
 				    wsi->user_space, NULL, 0))
 		goto bail2;
 
 	/* clear his proxy connection timeout */
-
 	lws_set_timeout(wsi, NO_PENDING_TIMEOUT, 0);
 
 	/* free up his parsing allocations */
-
-	lws_free(wsi->u.hdr.ah);
+	lws_header_table_detach(wsi);
 
 	lws_union_transition(wsi, LWSCM_WS_CLIENT);
 	wsi->state = LWSS_ESTABLISHED;
@@ -741,13 +810,12 @@ check_accept:
 	 * size mentioned in the protocol definition.  If 0 there, then
 	 * use a big default for compatibility
 	 */
-
 	n = wsi->protocol->rx_buffer_size;
 	if (!n)
 		n = LWS_MAX_SOCKET_IO_BUF;
-	n += LWS_SEND_BUFFER_PRE_PADDING + LWS_SEND_BUFFER_POST_PADDING;
-	wsi->u.ws.rx_user_buffer = lws_malloc(n);
-	if (!wsi->u.ws.rx_user_buffer) {
+	n += LWS_PRE;
+	wsi->u.ws.rx_ubuf = lws_malloc(n + 4 /* 0x0000ffff zlib */);
+	if (!wsi->u.ws.rx_ubuf) {
 		lwsl_err("Out of Mem allocating rx buffer %d\n", n);
 		goto bail2;
 	}
@@ -772,17 +840,16 @@ check_accept:
 	 * inform all extensions, not just active ones since they
 	 * already know
 	 */
-
 	ext = context->extensions;
 
 	while (ext && ext->callback) {
 		v = NULL;
-		for (n = 0; n < wsi->count_active_extensions; n++)
+		for (n = 0; n < wsi->count_act_ext; n++)
 			if (wsi->active_extensions[n] == ext)
-				v = wsi->active_extensions_user[n];
+				v = wsi->act_ext_user[n];
 
 		ext->callback(context, ext, wsi,
-			  LWS_EXT_CALLBACK_ANY_WSI_ESTABLISHED, v, NULL, 0);
+			  LWS_EXT_CB_ANY_WSI_ESTABLISHED, v, NULL, 0);
 		ext++;
 	}
 #endif
@@ -793,7 +860,7 @@ bail3:
 	close_reason = LWS_CLOSE_STATUS_NOSTATUS;
 
 bail2:
-	if (wsi->protocol) {
+	if (wsi->protocol && wsi->state == LWSS_ESTABLISHED) {
 		if (isErrorCodeReceived && p) {
 			wsi->protocol->callback(wsi,
 				LWS_CALLBACK_CLIENT_CONNECTION_ERROR,
@@ -893,16 +960,14 @@ lws_generate_client_handshake(struct lws *wsi, char *pkt)
 #ifndef LWS_NO_EXTENSIONS
 	ext = context->extensions;
 	while (ext && ext->callback) {
-
 		n = lws_ext_cb_all_exts(context, wsi,
-			   LWS_EXT_CALLBACK_CHECK_OK_TO_PROPOSE_EXTENSION,
+			   LWS_EXT_CB_CHECK_OK_TO_PROPOSE_EXTENSION,
 			   (char *)ext->name, 0);
 		if (n) { /* an extension vetos us */
 			lwsl_ext("ext %s vetoed\n", (char *)ext->name);
 			ext++;
 			continue;
 		}
-
 		n = context->protocols[0].callback(wsi,
 			LWS_CALLBACK_CLIENT_CONFIRM_EXTENSION_SUPPORTED,
 				wsi->user_space, (char *)ext->name, 0);
@@ -923,7 +988,7 @@ lws_generate_client_handshake(struct lws *wsi, char *pkt)
 
 		if (ext_count)
 			*p++ = ',';
-		p += sprintf(p, "%s", ext->name);
+		p += sprintf(p, "%s", ext->client_offer);
 		ext_count++;
 
 		ext++;
@@ -937,10 +1002,8 @@ lws_generate_client_handshake(struct lws *wsi, char *pkt)
 
 	/* give userland a chance to append, eg, cookies */
 
-	context->protocols[0].callback(wsi,
-				       LWS_CALLBACK_CLIENT_APPEND_HANDSHAKE_HEADER,
-				       NULL, &p,
-				       (pkt + sizeof(context->serv_buf)) - p - 12);
+	context->protocols[0].callback(wsi, LWS_CALLBACK_CLIENT_APPEND_HANDSHAKE_HEADER,
+				       NULL, &p, (pkt + LWS_MAX_SOCKET_IO_BUF) - p - 12);
 
 	p += sprintf(p, "\x0d\x0a");
 
