@@ -1,7 +1,7 @@
 /*
  * libwebsockets - small server side websockets and web server implementation
  *
- * Copyright (C) 2010-2015 Andy Green <andy@warmcat.com>
+ * Copyright (C) 2010-2016 Andy Green <andy@warmcat.com>
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -22,8 +22,9 @@
 
 #include "private-libwebsockets.h"
 
-int lws_context_init_server(struct lws_context_creation_info *info,
-			    struct lws_context *context)
+int
+lws_context_init_server(struct lws_context_creation_info *info,
+			struct lws_context *context)
 {
 #ifdef LWS_USE_IPV6
 	struct sockaddr_in6 serv_addr6;
@@ -31,12 +32,13 @@ int lws_context_init_server(struct lws_context_creation_info *info,
 #if LWS_POSIX
 	struct sockaddr_in serv_addr4;
 	socklen_t len = sizeof(struct sockaddr);
+	int n, opt = 1, limit = 1;
 	struct sockaddr_in sin;
 	struct sockaddr *v;
-	int n, opt = 1;
 #endif
 	lws_sockfd_type sockfd;
 	struct lws *wsi;
+	int m = 0;
 
 	/* set up our external listening socket we serve on */
 
@@ -44,6 +46,11 @@ int lws_context_init_server(struct lws_context_creation_info *info,
 		return 0;
 
 #if LWS_POSIX
+#if defined(__linux__)
+	limit = context->count_threads;
+#endif
+
+	for (m = 0; m < limit; m++) {
 #ifdef LWS_USE_IPV6
 	if (LWS_IPV6_ENABLED(context))
 		sockfd = socket(AF_INET6, SOCK_STREAM, 0);
@@ -69,6 +76,14 @@ int lws_context_init_server(struct lws_context_creation_info *info,
 		compatible_close(sockfd);
 		return 1;
 	}
+#if defined(__linux__) && defined(SO_REUSEPORT) && LWS_MAX_SMP > 1
+	if (context->count_threads > 1)
+		if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEPORT,
+				(const void *)&opt, sizeof(opt)) < 0) {
+			compatible_close(sockfd);
+			return 1;
+		}
+#endif
 #endif
 	lws_plat_set_socket_options(context, sockfd);
 
@@ -111,7 +126,6 @@ int lws_context_init_server(struct lws_context_creation_info *info,
 	else
 		info->port = ntohs(sin.sin_port);
 #endif
-
 	context->listen_port = info->port;
 
 	wsi = lws_zalloc(sizeof(struct lws));
@@ -123,18 +137,20 @@ int lws_context_init_server(struct lws_context_creation_info *info,
 	wsi->sock = sockfd;
 	wsi->mode = LWSCM_SERVER_LISTENER;
 	wsi->protocol = context->protocols;
+	wsi->tsi = m;
 
+	context->pt[m].wsi_listening = wsi;
 	if (insert_wsi_socket_into_fds(context, wsi))
 		goto bail;
 
-	context->lserv_mod = LWS_lserv_mod;
-	context->lserv_count = 0;
-	context->lserv_fd = sockfd;
+	context->count_wsi_allocated++;
+	context->pt[m].lserv_fd = sockfd;
 
 #if LWS_POSIX
-	listen(sockfd, LWS_SOMAXCONN);
+	listen(wsi->sock, LWS_SOMAXCONN);
+	} /* for each thread able to independently lister */
 #else
-	mbed3_tcp_stream_bind(sockfd, info->port, wsi);
+	mbed3_tcp_stream_bind(wsi->sock, info->port, wsi);
 #endif
 	lwsl_notice(" Listening on port %d\n", info->port);
 
@@ -147,44 +163,28 @@ bail:
 }
 
 int
-_lws_rx_flow_control(struct lws *wsi)
+_lws_server_listen_accept_flow_control(struct lws *twsi, int on)
 {
-	if (!wsi || wsi->context->being_destroyed)
+	struct lws_context_per_thread *pt = &twsi->context->pt[(int)twsi->tsi];
+	struct lws *wsi = pt->wsi_listening;
+	int n;
+
+	if (!wsi || twsi->context->being_destroyed)
 		return 0;
 
-	/* there is no pending change */
-	if (!(wsi->rxflow_change_to & LWS_RXFLOW_PENDING_CHANGE))
-		return 0;
+	lwsl_debug("%s: Thr %d: LISTEN wsi %p: state %d\n",
+		   __func__, twsi->tsi, (void *)wsi, on);
 
-	/* stuff is still buffered, not ready to really accept new input */
-	if (wsi->rxflow_buffer) {
-		/* get ourselves called back to deal with stashed buffer */
-		lws_callback_on_writable(wsi);
-		return 0;
-	}
+	if (on)
+		n = lws_change_pollfd(wsi, 0, LWS_POLLIN);
+	else
+		n = lws_change_pollfd(wsi, LWS_POLLIN, 0);
 
-	/* pending is cleared, we can change rxflow state */
-
-	wsi->rxflow_change_to &= ~LWS_RXFLOW_PENDING_CHANGE;
-
-	lwsl_info("rxflow: wsi %p change_to %d\n", wsi,
-			      wsi->rxflow_change_to & LWS_RXFLOW_ALLOW);
-
-	/* adjust the pollfd for this wsi */
-
-	if (wsi->rxflow_change_to & LWS_RXFLOW_ALLOW) {
-		if (lws_change_pollfd(wsi, 0, LWS_POLLIN)) {
-			lwsl_info("%s: fail\n", __func__);
-			return -1;
-		}
-	} else
-		if (lws_change_pollfd(wsi, LWS_POLLIN, 0))
-			return -1;
-
-	return 0;
+	return n;
 }
 
-int lws_http_action(struct lws *wsi)
+int
+lws_http_action(struct lws *wsi)
 {
 	enum http_connection_type connection_type;
 	enum http_version request_version;
@@ -292,25 +292,20 @@ int lws_http_action(struct lws *wsi)
 
 	n = wsi->protocol->callback(wsi, LWS_CALLBACK_FILTER_HTTP_CONNECTION,
 				    wsi->user_space, uri_ptr, uri_len);
-
-	if (!n) {
-		/*
-		 * if there is content supposed to be coming,
-		 * put a timeout on it having arrived
-		 */
-		lws_set_timeout(wsi, PENDING_TIMEOUT_HTTP_CONTENT,
-				AWAITING_TIMEOUT);
-
-		n = wsi->protocol->callback(wsi, LWS_CALLBACK_HTTP,
-					    wsi->user_space, uri_ptr, uri_len);
-	}
-
-	/* now drop the header info we kept a pointer to */
-	lws_free_set_NULL(wsi->u.http.ah);
-
 	if (n) {
 		lwsl_info("LWS_CALLBACK_HTTP closing\n");
-		return 1; /* struct ah ptr already nuked */		}
+
+		return 1;
+	}
+	/*
+	 * if there is content supposed to be coming,
+	 * put a timeout on it having arrived
+	 */
+	lws_set_timeout(wsi, PENDING_TIMEOUT_HTTP_CONTENT,
+			wsi->context->timeout_secs);
+
+	n = wsi->protocol->callback(wsi, LWS_CALLBACK_HTTP,
+				    wsi->user_space, uri_ptr, uri_len);
 
 	/*
 	 * If we're not issuing a file, check for content_length or
@@ -328,25 +323,31 @@ int lws_http_action(struct lws *wsi)
 	return 0;
 
 bail_nuke_ah:
-	/* drop the header info */
-	lws_free_set_NULL(wsi->u.hdr.ah);
+	/* we're closing, losing some rx is OK */
+	wsi->u.hdr.ah->rxpos = wsi->u.hdr.ah->rxlen;
+	lws_header_table_detach(wsi);
 
 	return 1;
 }
 
 
-int lws_handshake_server(struct lws *wsi, unsigned char **buf, size_t len)
+int
+lws_handshake_server(struct lws *wsi, unsigned char **buf, size_t len)
 {
 	struct lws_context *context = lws_get_context(wsi);
+	struct lws_context_per_thread *pt = &context->pt[(int)wsi->tsi];
+	struct _lws_header_related hdr;
 	struct allocated_headers *ah;
 	int protocol_len, n, hit;
 	char protocol_list[128];
 	char protocol_name[32];
 	char *p;
 
-	/* LWSCM_WS_SERVING */
+	assert(len < 10000000);
+	assert(wsi->u.hdr.ah);
 
 	while (len--) {
+		wsi->u.hdr.more_rx_waiting = !!len;
 
 		assert(wsi->mode == LWSCM_HTTP_SERVING);
 
@@ -486,12 +487,9 @@ upgrade_ws:
 			lwsl_info("checking %s\n", protocol_name);
 
 			n = 0;
-			while (wsi->protocol && context->protocols[n].callback) {
-				if (!wsi->protocol->name) {
-					n++;
-					continue;
-				}
-				if (!strcmp(context->protocols[n].name,
+			while (context->protocols[n].callback) {
+				if (context->protocols[n].name &&
+				    !strcmp(context->protocols[n].name,
 					    protocol_name)) {
 					lwsl_info("prot match %d\n", n);
 					wsi->protocol = &context->protocols[n];
@@ -506,19 +504,18 @@ upgrade_ws:
 		/* we didn't find a protocol he wanted? */
 
 		if (!hit) {
-			if (!lws_hdr_simple_ptr(wsi, WSI_TOKEN_PROTOCOL)) {
-				/*
-				 * some clients only have one protocol and
-				 * do not sent the protocol list header...
-				 * allow it and match to protocol 0
-				 */
-				lwsl_info("defaulting to prot 0 handler\n");
-				wsi->protocol = &context->protocols[0];
-			} else {
-				lwsl_err("No protocol from list \"%s\" supported\n",
+			if (lws_hdr_simple_ptr(wsi, WSI_TOKEN_PROTOCOL)) {
+				lwsl_err("No protocol from \"%s\" supported\n",
 					 protocol_list);
 				goto bail_nuke_ah;
 			}
+			/*
+			 * some clients only have one protocol and
+			 * do not sent the protocol list header...
+			 * allow it and match to protocol 0
+			 */
+			lwsl_info("defaulting to prot 0 handler\n");
+			wsi->protocol = &context->protocols[0];
 		}
 
 		/* allocate wsi->user storage */
@@ -538,7 +535,6 @@ upgrade_ws:
 			goto bail_nuke_ah;
 		}
 
-
 		/*
 		 * Perform the handshake according to the protocol version the
 		 * client announced
@@ -555,14 +551,39 @@ upgrade_ws:
 
 		default:
 			lwsl_warn("Unknown client spec version %d\n",
-						       wsi->ietf_spec_revision);
+				  wsi->ietf_spec_revision);
 			goto bail_nuke_ah;
 		}
 
-		/* drop the header info -- no bail_nuke_ah after this */
-		lws_free_header_table(wsi);
+		/* we are upgrading to ws, so http/1.1 and keepalive +
+		 * pipelined header considerations about keeping the ah around
+		 * no longer apply.  However it's common for the first ws
+		 * protocol data to have been coalesced with the browser
+		 * upgrade request and to already be in the ah rx buffer.
+		 */
+
+		lwsl_info("%s: %p: inheriting ah in ws mode (rxpos:%d, rxlen:%d)\n",
+			  __func__, wsi, wsi->u.hdr.ah->rxpos,
+			  wsi->u.hdr.ah->rxlen);
+		lws_pt_lock(pt);
+		hdr = wsi->u.hdr;
 
 		lws_union_transition(wsi, LWSCM_WS_SERVING);
+		/*
+		 * first service is WS mode will notice this, use the RX and
+		 * then detach the ah (caution: we are not in u.hdr union
+		 * mode any more then... ah_temp member is at start the same
+		 * though)
+		 *
+		 * Because rxpos/rxlen shows something in the ah, we will get
+		 * service guaranteed next time around the event loop
+		 *
+		 * All union members begin with hdr, so we can use it even
+		 * though we transitioned to ws union mode (the ah detach
+		 * code uses it anyway).
+		 */
+		wsi->u.hdr = hdr;
+		lws_pt_unlock(pt);
 
 		/*
 		 * create the frame buffer for this connection according to the
@@ -573,9 +594,9 @@ upgrade_ws:
 		n = wsi->protocol->rx_buffer_size;
 		if (!n)
 			n = LWS_MAX_SOCKET_IO_BUF;
-		n += LWS_SEND_BUFFER_PRE_PADDING + LWS_SEND_BUFFER_POST_PADDING;
-		wsi->u.ws.rx_user_buffer = lws_malloc(n);
-		if (!wsi->u.ws.rx_user_buffer) {
+		n += LWS_PRE;
+		wsi->u.ws.rx_ubuf = lws_malloc(n + 4 /* 0x0000ffff zlib */);
+		if (!wsi->u.ws.rx_ubuf) {
 			lwsl_err("Out of Mem allocating rx buffer %d\n", n);
 			return 1;
 		}
@@ -597,14 +618,40 @@ upgrade_ws:
 
 bail_nuke_ah:
 	/* drop the header info */
-	lws_free_header_table(wsi);
+	/* we're closing, losing some rx is OK */
+	wsi->u.hdr.ah->rxpos = wsi->u.hdr.ah->rxlen;
+	lws_header_table_detach(wsi);
+
 	return 1;
 }
 
+static int
+lws_get_idlest_tsi(struct lws_context *context)
+{
+	unsigned int lowest = ~0;
+	int n = 0, hit = -1;
+
+	for (; n < context->count_threads; n++) {
+		if ((unsigned int)context->pt[n].fds_count != context->fd_limit_per_thread - 1 &&
+		    (unsigned int)context->pt[n].fds_count < lowest) {
+			lowest = context->pt[n].fds_count;
+			hit = n;
+		}
+	}
+
+	return hit;
+}
+
 struct lws *
 lws_create_new_server_wsi(struct lws_context *context)
 {
 	struct lws *new_wsi;
+	int n = lws_get_idlest_tsi(context);
+
+	if (n < 0) {
+		lwsl_err("no space for new conn\n");
+		return NULL;
+	}
 
 	new_wsi = lws_zalloc(sizeof(struct lws));
 	if (new_wsi == NULL) {
@@ -612,6 +659,9 @@ lws_create_new_server_wsi(struct lws_context *context)
 		return NULL;
 	}
 
+	new_wsi->tsi = n;
+	lwsl_info("Accepted %p to tsi %d\n", new_wsi, new_wsi->tsi);
+
 	new_wsi->context = context;
 	new_wsi->pending_timeout = NO_PENDING_TIMEOUT;
 	new_wsi->rxflow_change_to = LWS_RXFLOW_ALLOW;
@@ -626,11 +676,6 @@ lws_create_new_server_wsi(struct lws_context *context)
 	new_wsi->use_ssl = LWS_SSL_ENABLED(context);
 #endif
 
-	if (lws_allocate_header_table(new_wsi)) {
-		lws_free(new_wsi);
-		return NULL;
-	}
-
 	/*
 	 * these can only be set once the protocol is known
 	 * we set an unestablished connection's protocol pointer
@@ -641,13 +686,14 @@ lws_create_new_server_wsi(struct lws_context *context)
 	new_wsi->user_space = NULL;
 	new_wsi->ietf_spec_revision = 0;
 	new_wsi->sock = LWS_SOCK_INVALID;
+	context->count_wsi_allocated++;
 
 	/*
 	 * outermost create notification for wsi
 	 * no user_space because no protocol selection
 	 */
-	context->protocols[0].callback(new_wsi, LWS_CALLBACK_WSI_CREATE, NULL,
-				       NULL, 0);
+	context->protocols[0].callback(new_wsi, LWS_CALLBACK_WSI_CREATE,
+				       NULL, NULL, 0);
 
 	return new_wsi;
 }
@@ -661,12 +707,13 @@ lws_create_new_server_wsi(struct lws_context *context)
  *	  transaction if possible
  */
 
-LWS_VISIBLE
-int lws_http_transaction_completed(struct lws *wsi)
+LWS_VISIBLE int LWS_WARN_UNUSED_RESULT
+lws_http_transaction_completed(struct lws *wsi)
 {
+	lwsl_debug("%s: wsi %p\n", __func__, wsi);
 	/* if we can't go back to accept new headers, drop the connection */
 	if (wsi->u.http.connection_type != HTTP_CONNECTION_KEEP_ALIVE) {
-		lwsl_info("%s: close connection\n", __func__);
+		lwsl_info("%s: %p: close connection\n", __func__, wsi);
 		return 1;
 	}
 
@@ -674,31 +721,117 @@ int lws_http_transaction_completed(struct lws *wsi)
 	wsi->state = LWSS_HTTP;
 	wsi->mode = LWSCM_HTTP_SERVING;
 	wsi->u.http.content_length = 0;
+	wsi->hdr_parsing_completed = 0;
 
 	/* He asked for it to stay alive indefinitely */
 	lws_set_timeout(wsi, NO_PENDING_TIMEOUT, 0);
 
-	if (lws_allocate_header_table(wsi))
-		return 1;
+	/*
+	 * We already know we are on http1.1 / keepalive and the next thing
+	 * coming will be another header set.
+	 *
+	 * If there is no pending rx and we still have the ah, drop it and
+	 * reacquire a new ah when the new headers start to arrive.  (Otherwise
+	 * we needlessly hog an ah indefinitely.)
+	 *
+	 * However if there is pending rx and we know from the keepalive state
+	 * that is already at least the start of another header set, simply
+	 * reset the existing header table and keep it.
+	 */
+	if (wsi->u.hdr.ah) {
+		if (!wsi->u.hdr.more_rx_waiting) {
+			wsi->u.hdr.ah->rxpos = wsi->u.hdr.ah->rxlen;
+			lws_header_table_detach(wsi);
+		} else
+			lws_header_table_reset(wsi);
+	}
 
 	/* If we're (re)starting on headers, need other implied init */
 	wsi->u.hdr.ues = URIES_IDLE;
 
-	lwsl_info("%s: keep-alive await new transaction\n", __func__);
+	lwsl_info("%s: %p: keep-alive await new transaction\n", __func__, wsi);
 
 	return 0;
 }
 
-LWS_VISIBLE
-int lws_server_socket_service(struct lws_context *context,
-			      struct lws *wsi, struct lws_pollfd *pollfd)
+/**
+ * lws_adopt_socket() - adopt foreign socket as if listen socket accepted it
+ * @context: lws context
+ * @accept_fd: fd of already-accepted socket to adopt
+ *
+ * Either returns new wsi bound to accept_fd, or closes accept_fd and
+ * returns NULL, having cleaned up any new wsi pieces.
+ *
+ * LWS adopts the socket in http serving mode, it's ready to accept an upgrade
+ * to ws or just serve http.
+ */
+
+LWS_VISIBLE struct lws *
+lws_adopt_socket(struct lws_context *context, lws_sockfd_type accept_fd)
 {
+	struct lws *new_wsi = lws_create_new_server_wsi(context);
+
+	if (!new_wsi) {
+		compatible_close(accept_fd);
+		return NULL;
+	}
+
+	lwsl_debug("%s: new wsi %p\n", __func__, new_wsi);
+
+	new_wsi->sock = accept_fd;
+
+	/* the transport is accepted... give him time to negotiate */
+	lws_set_timeout(new_wsi, PENDING_TIMEOUT_ESTABLISH_WITH_SERVER,
+			context->timeout_secs);
+
+#if LWS_POSIX == 0
+	mbed3_tcp_stream_accept(accept_fd, new_wsi);
+#endif
+
+	/*
+	 * A new connection was accepted. Give the user a chance to
+	 * set properties of the newly created wsi. There's no protocol
+	 * selected yet so we issue this to protocols[0]
+	 */
+	if ((context->protocols[0].callback)(new_wsi,
+	     LWS_CALLBACK_SERVER_NEW_CLIENT_INSTANTIATED, NULL, NULL, 0)) {
+		compatible_close(new_wsi->sock);
+		lws_free(new_wsi);
+		return NULL;
+	}
+
+	lws_libev_accept(new_wsi, new_wsi->sock);
+	lws_libuv_accept(new_wsi, new_wsi->sock);
+
+	if (!LWS_SSL_ENABLED(context)) {
+		if (insert_wsi_socket_into_fds(context, new_wsi))
+			goto fail;
+	} else {
+		new_wsi->mode = LWSCM_SSL_INIT;
+		if (lws_server_socket_service_ssl(new_wsi, accept_fd))
+			goto fail;
+	}
+
+	return new_wsi;
+
+fail:
+	lwsl_err("%s: fail\n", __func__);
+	lws_close_free_wsi(new_wsi, LWS_CLOSE_STATUS_NOSTATUS);
+
+	return NULL;
+}
+
+LWS_VISIBLE int
+lws_server_socket_service(struct lws_context *context, struct lws *wsi,
+			  struct lws_pollfd *pollfd)
+{
+	struct lws_context_per_thread *pt = &context->pt[(int)wsi->tsi];
 	lws_sockfd_type accept_fd = LWS_SOCK_INVALID;
+	struct allocated_headers *ah;
 #if LWS_POSIX
 	struct sockaddr_in cli_addr;
 	socklen_t clilen;
 #endif
-	struct lws *new_wsi = NULL;
 	int n, len;
 
 	switch (wsi->mode) {
@@ -715,11 +848,12 @@ int lws_server_socket_service(struct lws_context *context,
 			if (!(pollfd->revents & LWS_POLLOUT))
 				break;
 
-			if (lws_issue_raw(wsi, wsi->trunc_alloc + wsi->trunc_offset,
+			if (lws_issue_raw(wsi, wsi->trunc_alloc +
+					       wsi->trunc_offset,
 					  wsi->trunc_len) < 0)
 				goto fail;
 			/*
-			 * we can't afford to allow input processing send
+			 * we can't afford to allow input processing to send
 			 * something new, so spin around he event loop until
 			 * he doesn't have any partials
 			 */
@@ -728,18 +862,76 @@ int lws_server_socket_service(struct lws_context *context,
 
 		/* any incoming data ready? */
 
-		if (!(pollfd->revents & LWS_POLLIN))
+		if (!(pollfd->revents & pollfd->events & LWS_POLLIN))
 			goto try_pollout;
 
-		len = lws_ssl_capable_read(wsi, context->serv_buf,
-					   sizeof(context->serv_buf));
-		lwsl_debug("%s: read %d\r\n", __func__, len);
+		/* these states imply we MUST have an ah attached */
+
+		if (wsi->state == LWSS_HTTP ||
+		    wsi->state == LWSS_HTTP_ISSUING_FILE ||
+		    wsi->state == LWSS_HTTP_HEADERS) {
+			if (!wsi->u.hdr.ah)
+				if (lws_header_table_attach(wsi))
+					goto try_pollout;
+
+			ah = wsi->u.hdr.ah;
+
+			lwsl_debug("%s: %p: rxpos:%d rxlen:%d\n", __func__, wsi,
+				   ah->rxpos, ah->rxlen);
+
+			/* if nothing in ah rx buffer, get some fresh rx */
+			if (ah->rxpos == ah->rxlen) {
+				ah->rxlen = lws_ssl_capable_read(wsi, ah->rx,
+						   sizeof(ah->rx));
+				ah->rxpos = 0;
+				lwsl_debug("%s: wsi %p, ah->rxlen = %d\r\n",
+					   __func__, wsi, ah->rxlen);
+				switch (ah->rxlen) {
+				case 0:
+					lwsl_info("%s: read 0 len\n", __func__);
+					/* lwsl_info("   state=%d\n", wsi->state); */
+//					if (!wsi->hdr_parsing_completed)
+//						lws_header_table_detach(wsi);
+					/* fallthru */
+				case LWS_SSL_CAPABLE_ERROR:
+					goto fail;
+				case LWS_SSL_CAPABLE_MORE_SERVICE:
+					ah->rxlen = ah->rxpos = 0;
+					goto try_pollout;
+				}
+			}
+			assert(ah->rxpos != ah->rxlen && ah->rxlen);
+			/* just ignore incoming if waiting for close */
+			if (wsi->state != LWSS_FLUSHING_STORED_SEND_BEFORE_CLOSE) {
+				n = lws_read(wsi, ah->rx + ah->rxpos,
+					     ah->rxlen - ah->rxpos);
+				if (n < 0) /* we closed wsi */
+					return 1;
+				if (wsi->u.hdr.ah) {
+					if ( wsi->u.hdr.ah->rxlen)
+						 wsi->u.hdr.ah->rxpos += n;
+
+					if (wsi->u.hdr.ah->rxpos == wsi->u.hdr.ah->rxlen &&
+					    (wsi->mode != LWSCM_HTTP_SERVING &&
+					     wsi->mode != LWSCM_HTTP_SERVING_ACCEPTED &&
+					     wsi->mode != LWSCM_HTTP2_SERVING))
+						lws_header_table_detach(wsi);
+				}
+				break;
+			}
+
+			goto try_pollout;
+		}
+
+		len = lws_ssl_capable_read(wsi, pt->serv_buf,
+					   LWS_MAX_SOCKET_IO_BUF);
+		lwsl_debug("%s: wsi %p read %d\r\n", __func__, wsi, len);
 		switch (len) {
 		case 0:
-			lwsl_info("lws_server_skt_srv: read 0 len\n");
+			lwsl_info("%s: read 0 len\n", __func__);
 			/* lwsl_info("   state=%d\n", wsi->state); */
-			if (!wsi->hdr_parsing_completed)
-				lws_free_header_table(wsi);
+//			if (!wsi->hdr_parsing_completed)
+//				lws_header_table_detach(wsi);
 			/* fallthru */
 		case LWS_SSL_CAPABLE_ERROR:
 			goto fail;
@@ -753,10 +945,9 @@ int lws_server_socket_service(struct lws_context *context,
 			 * hm this may want to send
 			 * (via HTTP callback for example)
 			 */
-			n = lws_read(wsi, context->serv_buf, len);
+			n = lws_read(wsi, pt->serv_buf, len);
 			if (n < 0) /* we closed wsi */
 				return 1;
-
 			/* hum he may have used up the
 			 * writability above */
 			break;
@@ -769,25 +960,28 @@ try_pollout:
 			break;
 
 		/* one shot */
-		if (lws_change_pollfd(wsi, LWS_POLLOUT, 0))
+		if (lws_change_pollfd(wsi, LWS_POLLOUT, 0)) {
+			lwsl_notice("%s a\n", __func__);
 			goto fail;
-
-		lws_libev_io(wsi, LWS_EV_STOP | LWS_EV_WRITE);
+		}
 
 		if (wsi->state != LWSS_HTTP_ISSUING_FILE) {
-			n = user_callback_handle_rxflow(
-					wsi->protocol->callback,
+			n = user_callback_handle_rxflow(wsi->protocol->callback,
 					wsi, LWS_CALLBACK_HTTP_WRITEABLE,
 					wsi->user_space, NULL, 0);
-			if (n < 0)
+			if (n < 0) {
+				lwsl_info("writeable_fail\n");
 				goto fail;
+			}
 			break;
 		}
 
 		/* >0 == completion, <0 == error */
 		n = lws_serve_http_file_fragment(wsi);
-		if (n < 0 || (n > 0 && lws_http_transaction_completed(wsi)))
+		if (n < 0 || (n > 0 && lws_http_transaction_completed(wsi))) {
+			lwsl_info("completed\n");
 			goto fail;
+		}
 		break;
 
 	case LWSCM_SERVER_LISTENER:
@@ -795,89 +989,65 @@ try_pollout:
 #if LWS_POSIX
 		/* pollin means a client has connected to us then */
 
-		if (!(pollfd->revents & LWS_POLLIN))
-			break;
+		do {
+			if (!(pollfd->revents & LWS_POLLIN) || !(pollfd->events & LWS_POLLIN))
+				break;
 
-		/* listen socket got an unencrypted connection... */
+			/* listen socket got an unencrypted connection... */
 
-		clilen = sizeof(cli_addr);
-		lws_latency_pre(context, wsi);
-		accept_fd  = accept(pollfd->fd, (struct sockaddr *)&cli_addr,
-				    &clilen);
-		lws_latency(context, wsi,
-			"unencrypted accept LWSCM_SERVER_LISTENER",
-						     accept_fd, accept_fd >= 0);
-		if (accept_fd < 0) {
-			if (LWS_ERRNO == LWS_EAGAIN ||
-			    LWS_ERRNO == LWS_EWOULDBLOCK) {
-				lwsl_debug("accept asks to try again\n");
+			clilen = sizeof(cli_addr);
+			lws_latency_pre(context, wsi);
+			accept_fd  = accept(pollfd->fd, (struct sockaddr *)&cli_addr,
+					    &clilen);
+			lws_latency(context, wsi, "listener accept", accept_fd,
+				    accept_fd >= 0);
+			if (accept_fd < 0) {
+				if (LWS_ERRNO == LWS_EAGAIN ||
+				    LWS_ERRNO == LWS_EWOULDBLOCK) {
+					lwsl_err("accept asks to try again\n");
+					break;
+				}
+				lwsl_err("ERROR on accept: %s\n", strerror(LWS_ERRNO));
 				break;
 			}
-			lwsl_warn("ERROR on accept: %s\n", strerror(LWS_ERRNO));
-			break;
-		}
 
-		lws_plat_set_socket_options(context, accept_fd);
-#else
-		/* not very beautiful... */
-		accept_fd = (lws_sockfd_type)pollfd;
-#endif
-		/*
-		 * look at who we connected to and give user code a chance
-		 * to reject based on client IP.  There's no protocol selected
-		 * yet so we issue this to protocols[0]
-		 */
+			lws_plat_set_socket_options(context, accept_fd);
 
-		if ((context->protocols[0].callback)(wsi,
-				LWS_CALLBACK_FILTER_NETWORK_CONNECTION,
-					   NULL, (void *)(long)accept_fd, 0)) {
-			lwsl_debug("Callback denied network connection\n");
-			compatible_close(accept_fd);
-			break;
-		}
-
-		new_wsi = lws_create_new_server_wsi(context);
-		if (new_wsi == NULL) {
-			compatible_close(accept_fd);
-			break;
-		}
-
-		new_wsi->sock = accept_fd;
-
-		/* the transport is accepted... give him time to negotiate */
-		lws_set_timeout(new_wsi, PENDING_TIMEOUT_ESTABLISH_WITH_SERVER,
-				AWAITING_TIMEOUT);
-
-#if LWS_POSIX == 0
-		mbed3_tcp_stream_accept(accept_fd, new_wsi);
-#endif
-
-		/*
-		 * A new connection was accepted. Give the user a chance to
-		 * set properties of the newly created wsi. There's no protocol
-		 * selected yet so we issue this to protocols[0]
-		 */
-		(context->protocols[0].callback)(new_wsi,
-			LWS_CALLBACK_SERVER_NEW_CLIENT_INSTANTIATED,
-			NULL, NULL, 0);
-
-		lws_libev_accept(new_wsi, accept_fd);
-
-		if (!LWS_SSL_ENABLED(context)) {
-#if LWS_POSIX
 			lwsl_debug("accepted new conn  port %u on fd=%d\n",
 					  ntohs(cli_addr.sin_port), accept_fd);
+
+#else
+			/* not very beautiful... */
+			accept_fd = (lws_sockfd_type)pollfd;
 #endif
-			if (insert_wsi_socket_into_fds(context, new_wsi))
-				goto fail;
-		}
-		break;
+			/*
+			 * look at who we connected to and give user code a chance
+			 * to reject based on client IP.  There's no protocol selected
+			 * yet so we issue this to protocols[0]
+			 */
+			if ((context->protocols[0].callback)(wsi,
+					LWS_CALLBACK_FILTER_NETWORK_CONNECTION,
+					NULL, (void *)(long)accept_fd, 0)) {
+				lwsl_debug("Callback denied network connection\n");
+				compatible_close(accept_fd);
+				break;
+			}
+
+			if (!lws_adopt_socket(context, accept_fd))
+				/* already closed cleanly as necessary */
+				return 1;
+
+#if LWS_POSIX
+		} while (pt->fds_count < context->fd_limit_per_thread - 1 &&
+			 lws_poll_listen_fd(&pt->fds[wsi->position_in_fds_table]) > 0);
+#endif
+		return 0;
 
 	default:
 		break;
 	}
 
-	if (!lws_server_socket_service_ssl(&wsi, new_wsi, accept_fd, pollfd))
+	if (!lws_server_socket_service_ssl(wsi, accept_fd))
 		return 0;
 
 fail:
@@ -905,17 +1075,15 @@ fail:
  *	the wsi should be left alone.
  */
 
-LWS_VISIBLE int lws_serve_http_file(struct lws *wsi, const char *file,
-				    const char *content_type,
-				    const char *other_headers,
-				    int other_headers_len)
+LWS_VISIBLE int
+lws_serve_http_file(struct lws *wsi, const char *file, const char *content_type,
+		    const char *other_headers, int other_headers_len)
 {
 	struct lws_context *context = lws_get_context(wsi);
-	unsigned char *response = context->serv_buf +
-				  LWS_SEND_BUFFER_PRE_PADDING;
+	struct lws_context_per_thread *pt = &context->pt[(int)wsi->tsi];
+	unsigned char *response = pt->serv_buf + LWS_PRE;
 	unsigned char *p = response;
-	unsigned char *end = p + sizeof(context->serv_buf) -
-			     LWS_SEND_BUFFER_PRE_PADDING;
+	unsigned char *end = p + LWS_MAX_SOCKET_IO_BUF - LWS_PRE;
 	int ret = 0;
 
 	wsi->u.http.fd = lws_plat_file_open(wsi, file, &wsi->u.http.filelen,
@@ -964,38 +1132,47 @@ LWS_VISIBLE int lws_serve_http_file(struct lws *wsi, const char *file,
 }
 
 int
-lws_interpret_incoming_packet(struct lws *wsi, unsigned char *buf, size_t len)
+lws_interpret_incoming_packet(struct lws *wsi, unsigned char **buf, size_t len)
 {
-	size_t n = 0;
 	int m;
 
+	lwsl_parser("%s: received %d byte packet\n", __func__, (int)len);
 #if 0
-	lwsl_parser("received %d byte packet\n", (int)len);
-	lwsl_hexdump(buf, len);
+	lwsl_hexdump(*buf, len);
 #endif
 
 	/* let the rx protocol state machine have as much as it needs */
 
-	while (n < len) {
+	while (len) {
 		/*
 		 * we were accepting input but now we stopped doing so
 		 */
 		if (!(wsi->rxflow_change_to & LWS_RXFLOW_ALLOW)) {
-			lws_rxflow_cache(wsi, buf, n, len);
-
+			lws_rxflow_cache(wsi, *buf, 0, len);
+			lwsl_parser("%s: cached %d\n", __func__, len);
 			return 1;
 		}
 
+		if (wsi->u.ws.rx_draining_ext) {
+			m = lws_rx_sm(wsi, 0);
+			if (m < 0)
+				return -1;
+			continue;
+		}
+
 		/* account for what we're using in rxflow buffer */
 		if (wsi->rxflow_buffer)
 			wsi->rxflow_pos++;
 
 		/* process the byte */
-		m = lws_rx_sm(wsi, buf[n++]);
+		m = lws_rx_sm(wsi, *(*buf)++);
 		if (m < 0)
 			return -1;
+		len--;
 	}
 
+	lwsl_parser("%s: exit with %d unused\n", __func__, (int)len);
+
 	return 0;
 }
 
@@ -1008,7 +1185,7 @@ lws_server_get_canonical_hostname(struct lws_context *context,
 #if LWS_POSIX
 	/* find canonical hostname */
 	gethostname((char *)context->canonical_hostname,
-				       sizeof(context->canonical_hostname) - 1);
+		    sizeof(context->canonical_hostname) - 1);
 
 	lwsl_notice(" canonical_hostname = %s\n", context->canonical_hostname);
 #else
