@@ -37,7 +37,7 @@ lws_calllback_as_writeable(struct lws *wsi)
 		n = LWS_CALLBACK_HTTP_WRITEABLE;
 		break;
 	}
-	lwsl_info("%s: %p (user=%p)\n", __func__, wsi, wsi->user_space);
+	lwsl_debug("%s: %p (user=%p)\n", __func__, wsi, wsi->user_space);
 	return user_callback_handle_rxflow(wsi->protocol->callback,
 					   wsi, (enum lws_callback_reasons) n,
 					   wsi->user_space, NULL, 0);
@@ -53,12 +53,18 @@ lws_handle_POLLOUT_event(struct lws *wsi, struct lws_pollfd *pollfd)
 #endif
 	int ret, m, n;
 
-	/* pending truncated sends have uber priority */
+	/*
+	 * user callback is lowest priority to get these notifications
+	 * actually, since other pending things cannot be disordered
+	 */
 
+	/* Priority 1: pending truncated sends are incomplete ws fragments
+	 *	       If anything else sent first the protocol would be
+	 *	       corrupted.
+	 */
 	if (wsi->trunc_len) {
-		if (lws_issue_raw(wsi, wsi->trunc_alloc +
-				wsi->trunc_offset,
-						wsi->trunc_len) < 0) {
+		if (lws_issue_raw(wsi, wsi->trunc_alloc + wsi->trunc_offset,
+				  wsi->trunc_len) < 0) {
 			lwsl_info("%s signalling to close\n", __func__);
 			return -1;
 		}
@@ -68,8 +74,10 @@ lws_handle_POLLOUT_event(struct lws *wsi, struct lws_pollfd *pollfd)
 		if (wsi->state == LWSS_FLUSHING_STORED_SEND_BEFORE_CLOSE)
 			return -1; /* retry closing now */
 
+
 #ifdef LWS_USE_HTTP2
-	/* protocol packets are next */
+	/* Priority 2: protocol packets
+	 */
 	if (wsi->pps) {
 		lwsl_info("servicing pps %d\n", wsi->pps);
 		switch (wsi->pps) {
@@ -86,8 +94,8 @@ lws_handle_POLLOUT_event(struct lws *wsi, struct lws_pollfd *pollfd)
 		return 0; /* leave POLLOUT active */
 	}
 #endif
-	/* pending control packets have next priority */
-
+	/* Priority 3: pending control packets (pong or close)
+	 */
 	if ((wsi->state == LWSS_ESTABLISHED &&
 	     wsi->u.ws.ping_pending_flag) ||
 	    (wsi->state == LWSS_RETURNED_CLOSE_ALREADY &&
@@ -96,8 +104,7 @@ lws_handle_POLLOUT_event(struct lws *wsi, struct lws_pollfd *pollfd)
 		if (wsi->u.ws.payload_is_close)
 			write_type = LWS_WRITE_CLOSE;
 
-		n = lws_write(wsi, &wsi->u.ws.ping_payload_buf[
-					LWS_SEND_BUFFER_PRE_PADDING],
+		n = lws_write(wsi, &wsi->u.ws.ping_payload_buf[LWS_PRE],
 			      wsi->u.ws.ping_payload_len, write_type);
 		if (n < 0)
 			return -1;
@@ -112,15 +119,30 @@ lws_handle_POLLOUT_event(struct lws *wsi, struct lws_pollfd *pollfd)
 		return 0;
 	}
 
-	/* if we are closing, don't confuse the user with writeable cb */
-
+	/* Priority 4: if we are closing, not allowed to send more data frags
+	 *	       which means user callback or tx ext flush banned now
+	 */
 	if (wsi->state == LWSS_RETURNED_CLOSE_ALREADY)
 		goto user_service;
 
-	/* if nothing critical, user can get the callback */
+	/* Priority 5: Tx path extension with more to send
+	 *
+	 *	       These are handled as new fragments each time around
+	 *	       So while we must block new writeable callback to enforce
+	 *	       payload ordering, but since they are always complete
+	 *	       fragments control packets can interleave OK.
+	 */
+	if (wsi->state == LWSS_ESTABLISHED && wsi->u.ws.tx_draining_ext) {
+		lwsl_ext("SERVICING TX EXT DRAINING\n");
+		if (lws_write(wsi, NULL, 0, LWS_WRITE_CONTINUATION) < 0)
+			return -1;
+		/* leave POLLOUT active */
+		return 0;
+	}
 
-	m = lws_ext_cb_wsi_active_exts(wsi, LWS_EXT_CALLBACK_IS_WRITEABLE,
-								       NULL, 0);
+	/* Priority 6: user can get the callback
+	 */
+	m = lws_ext_cb_active(wsi, LWS_EXT_CB_IS_WRITEABLE, NULL, 0);
 #ifndef LWS_NO_EXTENSIONS
 	if (!wsi->extension_data_pending)
 		goto user_service;
@@ -144,9 +166,9 @@ lws_handle_POLLOUT_event(struct lws *wsi, struct lws_pollfd *pollfd)
 
 		/* give every extension a chance to spill */
 
-		m = lws_ext_cb_wsi_active_exts(wsi,
-					LWS_EXT_CALLBACK_PACKET_TX_PRESEND,
-							           &eff_buf, 0);
+		m = lws_ext_cb_active(wsi,
+					LWS_EXT_CB_PACKET_TX_PRESEND,
+					       &eff_buf, 0);
 		if (m < 0) {
 			lwsl_err("ext reports fatal error\n");
 			return -1;
@@ -162,7 +184,7 @@ lws_handle_POLLOUT_event(struct lws *wsi, struct lws_pollfd *pollfd)
 
 		if (eff_buf.token_len) {
 			n = lws_issue_raw(wsi, (unsigned char *)eff_buf.token,
-							     eff_buf.token_len);
+					  eff_buf.token_len);
 			if (n < 0) {
 				lwsl_info("closing from POLLOUT spill\n");
 				return -1;
@@ -209,15 +231,12 @@ lws_handle_POLLOUT_event(struct lws *wsi, struct lws_pollfd *pollfd)
 user_service:
 	/* one shot */
 
-	if (pollfd) {
+	if (pollfd)
 		if (lws_change_pollfd(wsi, LWS_POLLOUT, 0)) {
 			lwsl_info("failed at set pollfd\n");
 			return 1;
 		}
 
-		lws_libev_io(wsi, LWS_EV_STOP | LWS_EV_WRITE);
-	}
-
 #ifdef LWS_USE_HTTP2
 	/*
 	 * we are the 'network wsi' for potentially many muxed child wsi with
@@ -272,12 +291,14 @@ notify:
 int
 lws_service_timeout_check(struct lws *wsi, unsigned int sec)
 {
+	struct lws_context_per_thread *pt = &wsi->context->pt[(int)wsi->tsi];
+	struct lws **pwsi;
+
 	/*
 	 * if extensions want in on it (eg, we are a mux parent)
 	 * give them a chance to service child timeouts
 	 */
-	if (lws_ext_cb_wsi_active_exts(wsi, LWS_EXT_CALLBACK_1HZ,
-					     NULL, sec) < 0)
+	if (lws_ext_cb_active(wsi, LWS_EXT_CB_1HZ, NULL, sec) < 0)
 		return 0;
 
 	if (!wsi->pending_timeout)
@@ -288,8 +309,28 @@ lws_service_timeout_check(struct lws *wsi, unsigned int sec)
 	 * connection
 	 */
 	if ((time_t)sec > wsi->pending_timeout_limit) {
-		lwsl_info("wsi %p: TIMEDOUT WAITING on %d\n",
-			  (void *)wsi, wsi->pending_timeout);
+#if LWS_POSIX
+		lwsl_notice("wsi %p: TIMEDOUT WAITING on %d (did hdr %d, ah %p, wl %d, pfd events %d)\n",
+			    (void *)wsi, wsi->pending_timeout,
+			    wsi->hdr_parsing_completed, wsi->u.hdr.ah,
+			    pt->ah_wait_list_length,
+			    pt->fds[wsi->sock].events);
+#endif
+		lws_pt_lock(pt);
+
+		pwsi = &pt->ah_wait_list;
+		if (!pwsi)
+			return 0;
+		while (*pwsi) {
+			if (*pwsi == wsi)
+				break;
+			pwsi = &(*pwsi)->u.hdr.ah_wait_list;
+			lwsl_err("%s: pwsi=%p\n", __func__, pwsi);
+		}
+		lws_pt_unlock(pt);
+
+		if (!*pwsi)
+			lwsl_err("*** not on ah wait list ***\n");
 		/*
 		 * Since he failed a timeout, he already had a chance to do
 		 * something and was unable to... that includes situations like
@@ -325,6 +366,124 @@ int lws_rxflow_cache(struct lws *wsi, unsigned char *buf, int n, int len)
 	return 0;
 }
 
+/* this is used by the platform service code to stop us waiting for network
+ * activity in poll() when we have something that already needs service
+ */
+
+int
+lws_service_adjust_timeout(struct lws_context *context, int timeout_ms, int tsi)
+{
+	struct lws_context_per_thread *pt = &context->pt[tsi];
+	int n;
+
+	/* Figure out if we really want to wait in poll()
+	 * We only need to wait if really nothing already to do and we have
+	 * to wait for something from network
+	 */
+
+	/* 1) if we know we are draining rx ext, do not wait in poll */
+	if (pt->rx_draining_ext_list)
+		timeout_ms = 0;
+
+#ifdef LWS_OPENSSL_SUPPORT
+	/* 2) if we know we have non-network pending data, do not wait in poll */
+	if (lws_ssl_anybody_has_buffered_read_tsi(context, tsi)) {
+		timeout_ms = 0;
+		lwsl_err("ssl buffered read\n");
+	}
+#endif
+
+	/* 3) if any ah has pending rx, do not wait in poll */
+	for (n = 0; n < context->max_http_header_pool; n++)
+		if (pt->ah_pool[n].rxpos != pt->ah_pool[n].rxlen) {
+			/* any ah with pending rx must be attached to someone */
+			assert(pt->ah_pool[n].wsi);
+			timeout_ms = 0;
+			break;
+		}
+
+	return timeout_ms;
+}
+
+/*
+ * guys that need POLLIN service again without waiting for network action
+ * can force POLLIN here if not flowcontrolled, so they will get service.
+ *
+ * Return nonzero if anybody got their POLLIN faked
+ */
+int
+lws_service_flag_pending(struct lws_context *context, int tsi)
+{
+	struct lws_context_per_thread *pt = &context->pt[tsi];
+#ifdef LWS_OPENSSL_SUPPORT
+	struct lws *wsi_next;
+#endif
+	struct lws *wsi;
+	int forced = 0;
+	int n;
+
+	/* POLLIN faking */
+
+	/*
+	 * 1) For all guys with already-available ext data to drain, if they are
+	 * not flowcontrolled, fake their POLLIN status
+	 */
+	wsi = pt->rx_draining_ext_list;
+	while (wsi) {
+		pt->fds[wsi->position_in_fds_table].revents |=
+			pt->fds[wsi->position_in_fds_table].events & LWS_POLLIN;
+		if (pt->fds[wsi->position_in_fds_table].revents &
+		    LWS_POLLIN)
+			forced = 1;
+		wsi = wsi->u.ws.rx_draining_ext_list;
+	}
+
+#ifdef LWS_OPENSSL_SUPPORT
+	/*
+	 * 2) For all guys with buffered SSL read data already saved up, if they
+	 * are not flowcontrolled, fake their POLLIN status so they'll get
+	 * service to use up the buffered incoming data, even though their
+	 * network socket may have nothing
+	 */
+	wsi = pt->pending_read_list;
+	while (wsi) {
+		wsi_next = wsi->pending_read_list_next;
+		pt->fds[wsi->position_in_fds_table].revents |=
+			pt->fds[wsi->position_in_fds_table].events & LWS_POLLIN;
+		if (pt->fds[wsi->position_in_fds_table].revents & LWS_POLLIN) {
+			forced = 1;
+			/*
+			 * he's going to get serviced now, take him off the
+			 * list of guys with buffered SSL.  If he still has some
+			 * at the end of the service, he'll get put back on the
+			 * list then.
+			 */
+			lws_ssl_remove_wsi_from_buffered_list(wsi);
+		}
+
+		wsi = wsi_next;
+	}
+#endif
+	/*
+	 * 3) For any wsi who have an ah with pending RX who did not
+	 * complete their current headers, and are not flowcontrolled,
+	 * fake their POLLIN status so they will be able to drain the
+	 * rx buffered in the ah
+	 */
+	for (n = 0; n < context->max_http_header_pool; n++)
+		if (pt->ah_pool[n].rxpos != pt->ah_pool[n].rxlen &&
+		    !pt->ah_pool[n].wsi->hdr_parsing_completed) {
+			pt->fds[pt->ah_pool[n].wsi->position_in_fds_table].revents |=
+				pt->fds[pt->ah_pool[n].wsi->position_in_fds_table].events &
+					LWS_POLLIN;
+			if (pt->fds[pt->ah_pool[n].wsi->position_in_fds_table].revents &
+			    LWS_POLLIN)
+				forced = 1;
+		}
+
+	return forced;
+}
+
 /**
  * lws_service_fd() - Service polled socket with something waiting
  * @context:	Websocket context
@@ -349,27 +508,20 @@ int lws_rxflow_cache(struct lws *wsi, unsigned char *buf, int n, int len)
  */
 
 LWS_VISIBLE int
-lws_service_fd(struct lws_context *context, struct lws_pollfd *pollfd)
+lws_service_fd_tsi(struct lws_context *context, struct lws_pollfd *pollfd, int tsi)
 {
-#if LWS_POSIX
-	int idx = 0;
-#endif
-	lws_sockfd_type our_fd = 0;
+	struct lws_context_per_thread *pt = &context->pt[tsi];
+	lws_sockfd_type our_fd = 0, tmp_fd;
 	struct lws_tokens eff_buf;
 	unsigned int pending = 0;
+	struct lws *wsi, *wsi1;
 	char draining_flow = 0;
-	lws_sockfd_type mfd;
 	int timed_out = 0;
-	struct lws *wsi;
 	time_t now;
 	int n, m;
 	int more;
 
-#if LWS_POSIX
-	if (context->lserv_fd)
-		idx = wsi_from_fd(context, context->lserv_fd)->position_in_fds_table;
-#endif
-         /*
+	/*
 	 * you can call us with pollfd = NULL to just allow the once-per-second
 	 * global timeout checks; if less than a second since the last check
 	 * it returns immediately then.
@@ -388,19 +540,32 @@ lws_service_fd(struct lws_context *context, struct lws_pollfd *pollfd)
 		if (pollfd)
 			our_fd = pollfd->fd;
 
-		for (n = 0; n < context->fds_count; n++) {
-			mfd = context->fds[n].fd;
-			wsi = wsi_from_fd(context, mfd);
-			if (!wsi)
-				continue;
-
-			if (lws_service_timeout_check(wsi, (unsigned int)now))
+		wsi = context->pt[tsi].timeout_list;
+		while (wsi) {
+			/* we have to take copies, because he may be deleted */
+			wsi1 = wsi->timeout_list;
+			tmp_fd = wsi->sock;
+			if (lws_service_timeout_check(wsi, (unsigned int)now)) {
 				/* he did time out... */
-				if (mfd == our_fd)
+				if (tmp_fd == our_fd)
 					/* it was the guy we came to service! */
 					timed_out = 1;
 					/* he's gone, no need to mark as handled */
+			}
+			wsi = wsi1;
 		}
+#if 0
+		{
+			char s[300], *p = s;
+
+			for (n = 0; n < context->count_threads; n++)
+				p += sprintf(p, " %7lu (%5d), ",
+					     context->pt[n].count_conns,
+					     context->pt[n].fds_count);
+
+			lwsl_notice("load: %s\n", s);
+		}
+#endif
 	}
 
 	/* the socket we came to service timed out, nothing to do */
@@ -423,46 +588,12 @@ lws_service_fd(struct lws_context *context, struct lws_pollfd *pollfd)
 	 */
 
 #if LWS_POSIX
-	/*
-	 * deal with listen service piggybacking
-	 * every lserv_mod services of other fds, we
-	 * sneak one in to service the listen socket if there's anything waiting
-	 *
-	 * To handle connection storms, as found in ab, if we previously saw a
-	 * pending connection here, it causes us to check again next time.
-	 */
-
-	if (context->lserv_fd && pollfd != &context->fds[idx]) {
-		context->lserv_count++;
-		if (context->lserv_seen ||
-		    context->lserv_count == context->lserv_mod) {
-			context->lserv_count = 0;
-			m = 1;
-			if (context->lserv_seen > 5)
-				m = 2;
-			while (m--) {
-				/*
-				 * even with extpoll, we prepared this
-				 * internal fds for listen
-				 */
-				n = lws_poll_listen_fd(&context->fds[idx]);
-				if (n <= 0) {
-					if (context->lserv_seen)
-						context->lserv_seen--;
-					break;
-				}
-				/* there's a conn waiting for us */
-				lws_service_fd(context, &context->fds[idx]);
-				context->lserv_seen++;
-			}
-		}
-	}
 
 	/* handle session socket closed */
 
-	if ((!(pollfd->revents & LWS_POLLIN)) &&
-			(pollfd->revents & LWS_POLLHUP)) {
-
+	if ((!(pollfd->revents & pollfd->events & LWS_POLLIN)) &&
+	    (pollfd->revents & LWS_POLLHUP)) {
+		wsi->socket_is_permanently_unusable = 1;
 		lwsl_debug("Session Socket %p (fd=%d) dead\n",
 						       (void *)wsi, pollfd->fd);
 
@@ -495,7 +626,7 @@ lws_service_fd(struct lws_context *context, struct lws_pollfd *pollfd)
 	case LWSCM_WS_CLIENT:
 	case LWSCM_HTTP2_SERVING:
 
-		/* the guy requested a callback when it was OK to write */
+		/* 1: something requested a callback when it was OK to write */
 
 		if ((pollfd->revents & LWS_POLLOUT) &&
 		    (wsi->state == LWSS_ESTABLISHED ||
@@ -503,16 +634,77 @@ lws_service_fd(struct lws_context *context, struct lws_pollfd *pollfd)
 		     wsi->state == LWSS_HTTP2_ESTABLISHED_PRE_SETTINGS ||
 		     wsi->state == LWSS_RETURNED_CLOSE_ALREADY ||
 		     wsi->state == LWSS_FLUSHING_STORED_SEND_BEFORE_CLOSE) &&
-			   lws_handle_POLLOUT_event(wsi, pollfd)) {
+		    lws_handle_POLLOUT_event(wsi, pollfd)) {
 			if (wsi->state == LWSS_RETURNED_CLOSE_ALREADY)
 				wsi->state = LWSS_FLUSHING_STORED_SEND_BEFORE_CLOSE;
 			lwsl_info("lws_service_fd: closing\n");
 			goto close_and_handled;
 		}
 
-		if (wsi->rxflow_buffer &&
-			      (wsi->rxflow_change_to & LWS_RXFLOW_ALLOW)) {
-			lwsl_info("draining rxflow\n");
+		if (wsi->state == LWSS_RETURNED_CLOSE_ALREADY ||
+		    wsi->state == LWSS_AWAITING_CLOSE_ACK) {
+			/*
+			 * we stopped caring about anything except control
+			 * packets.  Force flow control off, defeat tx
+			 * draining.
+			 */
+			lws_rx_flow_control(wsi, 1);
+			wsi->u.ws.tx_draining_ext = 0;
+		}
+
+		if (wsi->u.ws.tx_draining_ext) {
+			/* we cannot deal with new RX until the TX ext
+			 * path has been drained.  It's because new
+			 * rx will, eg, crap on the wsi rx buf that
+			 * may be needed to retain state.
+			 *
+			 * TX ext drain path MUST go through event loop
+			 * to avoid blocking.
+			 */
+			break;
+		}
+
+		if (!(wsi->rxflow_change_to & LWS_RXFLOW_ALLOW))
+			/* We cannot deal with any kind of new RX
+			 * because we are RX-flowcontrolled.
+			 */
+			break;
+
+		/* 2: RX Extension needs to be drained
+		 */
+
+		if (wsi->state == LWSS_ESTABLISHED &&
+		    wsi->u.ws.rx_draining_ext) {
+
+			lwsl_ext("%s: RX EXT DRAINING: Service\n", __func__);
+#ifndef LWS_NO_CLIENT
+			if (wsi->mode == LWSCM_WS_CLIENT) {
+				n = lws_client_rx_sm(wsi, 0);
+				if (n < 0)
+					/* we closed wsi */
+					n = 0;
+			} else
+#endif
+				n = lws_rx_sm(wsi, 0);
+
+			goto handled;
+		}
+
+		if (wsi->u.ws.rx_draining_ext)
+			/*
+			 * We have RX EXT content to drain, but can't do it
+			 * right now.  That means we cannot do anything lower
+			 * priority either.
+			 */
+			break;
+
+		/* 3: RX Flowcontrol buffer needs to be drained
+		 */
+
+		if (wsi->rxflow_buffer) {
+			lwsl_info("draining rxflow (len %d)\n",
+				wsi->rxflow_len - wsi->rxflow_pos
+			);
 			/* well, drain it */
 			eff_buf.token = (char *)wsi->rxflow_buffer +
 						wsi->rxflow_pos;
@@ -521,29 +713,42 @@ lws_service_fd(struct lws_context *context, struct lws_pollfd *pollfd)
 			goto drain;
 		}
 
-		/* any incoming data ready? */
+		/* 4: any incoming (or ah-stashed incoming rx) data ready?
+		 * notice if rx flow going off raced poll(), rx flow wins
+		 */
 
-		if (!(pollfd->revents & LWS_POLLIN))
+		if (!(pollfd->revents & pollfd->events & LWS_POLLIN))
 			break;
 read:
 
-		eff_buf.token_len = lws_ssl_capable_read(wsi,
-					context->serv_buf,
-					pending ? pending :
-					sizeof(context->serv_buf));
-		switch (eff_buf.token_len) {
-		case 0:
-			lwsl_info("service_fd: closing due to 0 length read\n");
-			goto close_and_handled;
-		case LWS_SSL_CAPABLE_MORE_SERVICE:
-			lwsl_info("SSL Capable more service\n");
-			n = 0;
-			goto handled;
-		case LWS_SSL_CAPABLE_ERROR:
-			lwsl_info("Closing when error\n");
-			goto close_and_handled;
-		}
+		/* all the union members start with hdr, so even in ws mode
+		 * we can deal with the ah via u.hdr
+		 */
+		if (wsi->u.hdr.ah) {
+			lwsl_err("%s: %p: using inherited ah rx\n", __func__, wsi);
+			eff_buf.token_len = wsi->u.hdr.ah->rxlen -
+					    wsi->u.hdr.ah->rxpos;
+			eff_buf.token = (char *)wsi->u.hdr.ah->rx +
+					wsi->u.hdr.ah->rxpos;
+		} else {
 
+			eff_buf.token_len = lws_ssl_capable_read(wsi, pt->serv_buf,
+					     pending ? pending : LWS_MAX_SOCKET_IO_BUF);
+			switch (eff_buf.token_len) {
+			case 0:
+				lwsl_info("service_fd: closing due to 0 length read\n");
+				goto close_and_handled;
+			case LWS_SSL_CAPABLE_MORE_SERVICE:
+				lwsl_info("SSL Capable more service\n");
+				n = 0;
+				goto handled;
+			case LWS_SSL_CAPABLE_ERROR:
+				lwsl_info("Closing when error\n");
+				goto close_and_handled;
+			}
+
+			eff_buf.token = (char *)pt->serv_buf;
+		}
 		/*
 		 * give any active extensions a chance to munge the buffer
 		 * before parse.  We pass in a pointer to an lws_tokens struct
@@ -555,15 +760,12 @@ read:
 		 * extension callback handling, just the normal input buffer is
 		 * used then so it is efficient.
 		 */
-
-		eff_buf.token = (char *)context->serv_buf;
 drain:
-
 		do {
 			more = 0;
 
-			m = lws_ext_cb_wsi_active_exts(wsi,
-				LWS_EXT_CALLBACK_PACKET_RX_PREPARSE, &eff_buf, 0);
+			m = lws_ext_cb_active(wsi, LWS_EXT_CB_PACKET_RX_PREPARSE,
+					      &eff_buf, 0);
 			if (m < 0)
 				goto close_and_handled;
 			if (m)
@@ -572,6 +774,13 @@ drain:
 			/* service incoming data */
 
 			if (eff_buf.token_len) {
+				/*
+				 * if draining from rxflow buffer, not
+				 * critical to track what was used since at the
+				 * use it bumps wsi->rxflow_pos.  If we come
+				 * around again it will pick up from where it
+				 * left off.
+				 */
 				n = lws_read(wsi, (unsigned char *)eff_buf.token,
 					     eff_buf.token_len);
 				if (n < 0) {
@@ -585,16 +794,26 @@ drain:
 			eff_buf.token_len = 0;
 		} while (more);
 
+		if (wsi->u.hdr.ah) {
+			lwsl_err("%s: %p: detaching inherited used ah\n", __func__, wsi);
+			/* show we used all the pending rx up */
+			wsi->u.hdr.ah->rxpos = wsi->u.hdr.ah->rxlen;
+			/* we can run the normal ah detach flow despite
+			 * being in ws union mode, since all union members
+			 * start with hdr */
+			lws_header_table_detach(wsi);
+		}
+
 		pending = lws_ssl_pending(wsi);
 		if (pending) {
 handle_pending:
-			pending = pending > sizeof(context->serv_buf) ?
-				sizeof(context->serv_buf) : pending;
+			pending = pending > LWS_MAX_SOCKET_IO_BUF ?
+					LWS_MAX_SOCKET_IO_BUF : pending;
 			goto read;
 		}
 
 		if (draining_flow && wsi->rxflow_buffer &&
-				 wsi->rxflow_pos == wsi->rxflow_len) {
+		    wsi->rxflow_pos == wsi->rxflow_len) {
 			lwsl_info("flow buffer: drained\n");
 			lws_free_set_NULL(wsi->rxflow_buffer);
 			/* having drained the rxflow buffer, can rearm POLLIN */
@@ -612,6 +831,8 @@ handle_pending:
 		break;
 #else
 		n = lws_client_socket_service(context, wsi, pollfd);
+		if (n)
+			return 1;
 		goto handled;
 #endif
 	}
@@ -634,6 +855,12 @@ handled:
 	return n;
 }
 
+LWS_VISIBLE int
+lws_service_fd(struct lws_context *context, struct lws_pollfd *pollfd)
+{
+	return lws_service_fd_tsi(context, pollfd, 0);
+}
+
 /**
  * lws_service() - Service any pending websocket activity
  * @context:	Websocket context
@@ -672,3 +899,9 @@ lws_service(struct lws_context *context, int timeout_ms)
 	return lws_plat_service(context, timeout_ms);
 }
 
+LWS_VISIBLE int
+lws_service_tsi(struct lws_context *context, int timeout_ms, int tsi)
+{
+	return lws_plat_service_tsi(context, timeout_ms, tsi);
+}
+
