@@ -24,6 +24,10 @@
  #include <openssl/err.h>
 #endif
 
+#if OPENSSL_VERSION_NUMBER >= 0x0090800fL
+#include <openssl/ecdh.h>
+#endif
+
 int openssl_websocket_private_data_index;
 
 static int
@@ -84,6 +88,85 @@ OpenSSL_verify_callback(int preverify_ok, X509_STORE_CTX *x509_ctx)
 	return !n;
 }
 
+static int
+lws_context_ssl_init_ecdh(struct lws_context *context)
+{
+#ifdef LWS_SSL_SERVER_WITH_ECDH_CERT
+	int KeyType;
+	EC_KEY *EC_key = NULL;
+	X509 *x;
+	EVP_PKEY *pkey;
+
+	if (!(context->options & LWS_SERVER_OPTION_SSL_ECDH))
+		return 0;
+
+	lwsl_notice(" Using ECDH certificate support\n");
+
+	/* Get X509 certificate from ssl context */
+	x = sk_X509_value(context->ssl_ctx->extra_certs, 0);
+	if (!x) {
+		lwsl_err("%s: x is NULL\n", __func__);
+		return 1;
+	}
+	/* Get the public key from certificate */
+	pkey = X509_get_pubkey(x);
+	if (!pkey) {
+		lwsl_err("%s: pkey is NULL\n", __func__);
+
+		return 1;
+	}
+	/* Get the key type */
+	KeyType = EVP_PKEY_type(pkey->type);
+
+	if (EVP_PKEY_EC != KeyType) {
+		lwsl_notice("Key type is not EC\n");
+		return 0;
+	}
+	/* Get the key */
+	EC_key = EVP_PKEY_get1_EC_KEY(pkey);
+	/* Set ECDH parameter */
+	if (!EC_key) {
+		lwsl_err("%s: ECDH key is NULL \n", __func__);
+		return 1;
+	}
+	SSL_CTX_set_tmp_ecdh(context->ssl_ctx, EC_key);
+	EC_KEY_free(EC_key);
+#endif
+	return 0;
+}
+
+static int
+lws_context_ssl_init_ecdh_curve(struct lws_context_creation_info *info,
+				struct lws_context *context)
+{
+#if OPENSSL_VERSION_NUMBER >= 0x0090800fL
+	EC_KEY *ecdh;
+	int ecdh_nid;
+	const char *ecdh_curve = "prime256v1";
+
+	ecdh_nid = OBJ_sn2nid(ecdh_curve);
+	if (NID_undef == ecdh_nid) {
+		lwsl_err("SSL: Unknown curve name '%s'", ecdh_curve);
+		return 1;
+	}
+
+	ecdh = EC_KEY_new_by_curve_name(ecdh_nid);
+	if (NULL == ecdh) {
+		lwsl_err("SSL: Unable to create curve '%s'", ecdh_curve);
+		return 1;
+	}
+	SSL_CTX_set_tmp_ecdh(context->ssl_ctx, ecdh);
+	EC_KEY_free(ecdh);
+
+	SSL_CTX_set_options(context->ssl_ctx, SSL_OP_SINGLE_ECDH_USE);
+
+	lwsl_notice(" SSL ECDH curve '%s'\n", ecdh_curve);
+#else
+	lwsl_notice(" OpenSSL doesn't support ECDH\n");
+#endif
+	return 0;
+}
+
 LWS_VISIBLE int
 lws_context_init_server_ssl(struct lws_context_creation_info *info,
 			    struct lws_context *context)
@@ -147,7 +230,7 @@ lws_context_init_server_ssl(struct lws_context_creation_info *info,
 		error = ERR_get_error();
 		lwsl_err("problem creating ssl method %lu: %s\n",
 			error, ERR_error_string(error,
-					      (char *)context->serv_buf));
+					      (char *)context->pt[0].serv_buf));
 		return 1;
 	}
 	context->ssl_ctx = SSL_CTX_new(method);	/* create context */
@@ -155,7 +238,7 @@ lws_context_init_server_ssl(struct lws_context_creation_info *info,
 		error = ERR_get_error();
 		lwsl_err("problem creating ssl context %lu: %s\n",
 			error, ERR_error_string(error,
-					      (char *)context->serv_buf));
+					      (char *)context->pt[0].serv_buf));
 		return 1;
 	}
 
@@ -164,6 +247,7 @@ lws_context_init_server_ssl(struct lws_context_creation_info *info,
 #ifdef SSL_OP_NO_COMPRESSION
 	SSL_CTX_set_options(context->ssl_ctx, SSL_OP_NO_COMPRESSION);
 #endif
+	SSL_CTX_set_options(context->ssl_ctx, SSL_OP_SINGLE_DH_USE);
 	SSL_CTX_set_options(context->ssl_ctx, SSL_OP_CIPHER_SERVER_PREFERENCE);
 	if (info->ssl_cipher_list)
 		SSL_CTX_set_cipher_list(context->ssl_ctx,
@@ -184,16 +268,25 @@ lws_context_init_server_ssl(struct lws_context_creation_info *info,
 
 		SSL_CTX_set_verify(context->ssl_ctx,
 		       verify_options, OpenSSL_verify_callback);
+	}
 
-		/*
-		 * give user code a chance to load certs into the server
-		 * allowing it to verify incoming client certs
-		 */
+	/*
+	 * give user code a chance to load certs into the server
+	 * allowing it to verify incoming client certs
+	 */
 
-		context->protocols[0].callback(&wsi,
+	if (info->ssl_ca_filepath &&
+	    !SSL_CTX_load_verify_locations(context->ssl_ctx,
+					   info->ssl_ca_filepath, NULL)) {
+		lwsl_err("%s: SSL_CTX_load_verify_locations unhappy\n");
+	}
+
+	if (lws_context_ssl_init_ecdh_curve(info, context))
+		return -1;
+
+	context->protocols[0].callback(&wsi,
 			LWS_CALLBACK_OPENSSL_LOAD_EXTRA_SERVER_VERIFY_CERTS,
 					       context->ssl_ctx, NULL, 0);
-	}
 
 	if (info->options & LWS_SERVER_OPTION_ALLOW_NON_SSL_ON_SSL_PORT)
 		/* Normally SSL listener rejects non-ssl, optionally allow */
@@ -211,7 +304,7 @@ lws_context_init_server_ssl(struct lws_context_creation_info *info,
 				info->ssl_cert_filepath,
 				error,
 				ERR_error_string(error,
-					      (char *)context->serv_buf));
+					      (char *)context->pt[0].serv_buf));
 			return 1;
 		}
 		lws_ssl_bind_passphrase(context->ssl_ctx, info);
@@ -225,7 +318,7 @@ lws_context_init_server_ssl(struct lws_context_creation_info *info,
 				lwsl_err("ssl problem getting key '%s' %lu: %s\n",
 					 info->ssl_private_key_filepath, error,
 					 ERR_error_string(error,
-					      (char *)context->serv_buf));
+					      (char *)context->pt[0].serv_buf));
 				return 1;
 			}
 		} else
@@ -243,6 +336,16 @@ lws_context_init_server_ssl(struct lws_context_creation_info *info,
 			return 1;
 		}
 
+		if (lws_context_ssl_init_ecdh(context))
+			return 1;
+
+#if OPENSSL_VERSION_NUMBER >= 0x10200000L
+		n = SSL_CTX_build_cert_chain((SSL_CTX *)user,
+					     SSL_BUILD_CHAIN_FLAG_CHECK |
+					     SSL_BUILD_CHAIN_FLAG_IGNORE_ERROR);
+		lwsl_notice("%s: build cert chain %d", __func__, n);
+#endif
+
 		/*
 		 * SSL is happy and has a cert it's content with
 		 * If we're supporting HTTP2, initialize that
@@ -318,7 +421,7 @@ int lws_context_init_client_ssl(struct lws_context_creation_info *info,
 		error = ERR_get_error();
 		lwsl_err("problem creating ssl method %lu: %s\n",
 			error, ERR_error_string(error,
-				      (char *)context->serv_buf));
+				      (char *)context->pt[0].serv_buf));
 		return 1;
 	}
 	/* create context */
@@ -327,7 +430,7 @@ int lws_context_init_client_ssl(struct lws_context_creation_info *info,
 		error = ERR_get_error();
 		lwsl_err("problem creating ssl context %lu: %s\n",
 			error, ERR_error_string(error,
-				      (char *)context->serv_buf));
+				      (char *)context->pt[0].serv_buf));
 		return 1;
 	}
 
@@ -382,7 +485,7 @@ int lws_context_init_client_ssl(struct lws_context_creation_info *info,
 				info->ssl_cert_filepath,
 				ERR_get_error(),
 				ERR_error_string(ERR_get_error(),
-				(char *)context->serv_buf));
+				(char *)context->pt[0].serv_buf));
 			return 1;
 		}
 	}
@@ -395,7 +498,7 @@ int lws_context_init_client_ssl(struct lws_context_creation_info *info,
 				info->ssl_private_key_filepath,
 				ERR_get_error(),
 				ERR_error_string(ERR_get_error(),
-				      (char *)context->serv_buf));
+				      (char *)context->pt[0].serv_buf));
 			return 1;
 		}
 
@@ -425,16 +528,17 @@ LWS_VISIBLE void
 lws_ssl_remove_wsi_from_buffered_list(struct lws *wsi)
 {
 	struct lws_context *context = wsi->context;
+	struct lws_context_per_thread *pt = &context->pt[(int)wsi->tsi];
 
 	if (!wsi->pending_read_list_prev &&
 	    !wsi->pending_read_list_next &&
-	    context->pending_read_list != wsi)
+	    pt->pending_read_list != wsi)
 		/* we are not on the list */
 		return;
 
 	/* point previous guy's next to our next */
 	if (!wsi->pending_read_list_prev)
-		context->pending_read_list = wsi->pending_read_list_next;
+		pt->pending_read_list = wsi->pending_read_list_next;
 	else
 		wsi->pending_read_list_prev->pending_read_list_next =
 			wsi->pending_read_list_next;
@@ -452,6 +556,7 @@ LWS_VISIBLE int
 lws_ssl_capable_read(struct lws *wsi, unsigned char *buf, int len)
 {
 	struct lws_context *context = wsi->context;
+	struct lws_context_per_thread *pt = &context->pt[(int)wsi->tsi];
 	int n;
 
 	if (!wsi->ssl)
@@ -487,16 +592,16 @@ lws_ssl_capable_read(struct lws *wsi, unsigned char *buf, int len)
 		return n;
 	if (wsi->pending_read_list_prev)
 		return n;
-	if (context->pending_read_list == wsi)
+	if (pt->pending_read_list == wsi)
 		return n;
 
 	/* add us to the linked list of guys with pending ssl */
-	if (context->pending_read_list)
-		context->pending_read_list->pending_read_list_prev = wsi;
+	if (pt->pending_read_list)
+		pt->pending_read_list->pending_read_list_prev = wsi;
 
-	wsi->pending_read_list_next = context->pending_read_list;
+	wsi->pending_read_list_next = pt->pending_read_list;
 	wsi->pending_read_list_prev = NULL;
-	context->pending_read_list = wsi;
+	pt->pending_read_list = wsi;
 
 	return n;
 bail:
@@ -556,12 +661,10 @@ lws_ssl_close(struct lws *wsi)
 /* leave all wsi close processing to the caller */
 
 LWS_VISIBLE int
-lws_server_socket_service_ssl(struct lws **pwsi, struct lws *new_wsi,
-			      lws_sockfd_type accept_fd,
-			      struct lws_pollfd *pollfd)
+lws_server_socket_service_ssl(struct lws *wsi, lws_sockfd_type accept_fd)
 {
-	struct lws *wsi = *pwsi;
 	struct lws_context *context = wsi->context;
+	struct lws_context_per_thread *pt = &context->pt[(int)wsi->tsi];
 	int n, m;
 #ifndef USE_WOLFSSL
 	BIO *bio;
@@ -571,41 +674,37 @@ lws_server_socket_service_ssl(struct lws **pwsi, struct lws *new_wsi,
 		return 0;
 
 	switch (wsi->mode) {
-	case LWSCM_SERVER_LISTENER:
+	case LWSCM_SSL_INIT:
 
-		if (!new_wsi) {
-			lwsl_err("no new_wsi\n");
-			return 0;
-		}
-
-		new_wsi->ssl = SSL_new(context->ssl_ctx);
-		if (new_wsi->ssl == NULL) {
+		wsi->ssl = SSL_new(context->ssl_ctx);
+		if (wsi->ssl == NULL) {
 			lwsl_err("SSL_new failed: %s\n",
-				 ERR_error_string(SSL_get_error(new_wsi->ssl, 0), NULL));
+				 ERR_error_string(SSL_get_error(wsi->ssl, 0), NULL));
 			lws_decode_ssl_error();
-			compatible_close(accept_fd);
+			if (accept_fd != LWS_SOCK_INVALID)
+				compatible_close(accept_fd);
 			goto fail;
 		}
 
-		SSL_set_ex_data(new_wsi->ssl,
+		SSL_set_ex_data(wsi->ssl,
 			openssl_websocket_private_data_index, context);
 
-		SSL_set_fd(new_wsi->ssl, accept_fd);
+		SSL_set_fd(wsi->ssl, accept_fd);
 
 #ifdef USE_WOLFSSL
 #ifdef USE_OLD_CYASSL
-		CyaSSL_set_using_nonblock(new_wsi->ssl, 1);
+		CyaSSL_set_using_nonblock(wsi->ssl, 1);
 #else
-		wolfSSL_set_using_nonblock(new_wsi->ssl, 1);
+		wolfSSL_set_using_nonblock(wsi->ssl, 1);
 #endif
 #else
-		SSL_set_mode(new_wsi->ssl, SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER);
-		bio = SSL_get_rbio(new_wsi->ssl);
+		SSL_set_mode(wsi->ssl, SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER);
+		bio = SSL_get_rbio(wsi->ssl);
 		if (bio)
 			BIO_set_nbio(bio, 1); /* nonblocking */
 		else
 			lwsl_notice("NULL rbio\n");
-		bio = SSL_get_wbio(new_wsi->ssl);
+		bio = SSL_get_wbio(wsi->ssl);
 		if (bio)
 			BIO_set_nbio(bio, 1); /* nonblocking */
 		else
@@ -618,14 +717,12 @@ lws_server_socket_service_ssl(struct lws **pwsi, struct lws *new_wsi,
 		 * pieces come if we're not sorted yet
 		 */
 
-		*pwsi = new_wsi;
-		wsi = *pwsi;
 		wsi->mode = LWSCM_SSL_ACK_PENDING;
 		if (insert_wsi_socket_into_fds(context, wsi))
 			goto fail;
 
 		lws_set_timeout(wsi, PENDING_TIMEOUT_SSL_ACCEPT,
-							AWAITING_TIMEOUT);
+				context->timeout_secs);
 
 		lwsl_info("inserted SSL accept into fds, trying SSL_accept\n");
 
@@ -636,12 +733,10 @@ lws_server_socket_service_ssl(struct lws **pwsi, struct lws *new_wsi,
 		if (lws_change_pollfd(wsi, LWS_POLLOUT, 0))
 			goto fail;
 
-		lws_libev_io(wsi, LWS_EV_STOP | LWS_EV_WRITE);
-
 		lws_latency_pre(context, wsi);
 
-		n = recv(wsi->sock, (char *)context->serv_buf,
-			sizeof(context->serv_buf), MSG_PEEK);
+		n = recv(wsi->sock, (char *)pt->serv_buf, LWS_MAX_SOCKET_IO_BUF,
+			 MSG_PEEK);
 
 		/*
 		 * optionally allow non-SSL connect on SSL listening socket
@@ -651,7 +746,7 @@ lws_server_socket_service_ssl(struct lws **pwsi, struct lws *new_wsi,
 		 */
 
 		if (context->allow_non_ssl_on_ssl_port) {
-			if (n >= 1 && context->serv_buf[0] >= ' ') {
+			if (n >= 1 && pt->serv_buf[0] >= ' ') {
 				/*
 				* TLS content-type for Handshake is 0x16, and
 				* for ChangeCipherSpec Record, it's 0x14
@@ -704,8 +799,6 @@ go_again:
 			if (lws_change_pollfd(wsi, 0, LWS_POLLIN))
 				goto fail;
 
-			lws_libev_io(wsi, LWS_EV_START | LWS_EV_READ);
-
 			lwsl_info("SSL_ERROR_WANT_READ\n");
 			break;
 		}
@@ -713,17 +806,16 @@ go_again:
 			if (lws_change_pollfd(wsi, 0, LWS_POLLOUT))
 				goto fail;
 
-			lws_libev_io(wsi, LWS_EV_START | LWS_EV_WRITE);
 			break;
 		}
 		lwsl_debug("SSL_accept failed skt %u: %s\n",
-			   pollfd->fd, ERR_error_string(m, NULL));
+			   wsi->sock, ERR_error_string(m, NULL));
 		goto fail;
 
 accepted:
 		/* OK, we are accepted... give him some time to negotiate */
 		lws_set_timeout(wsi, PENDING_TIMEOUT_ESTABLISH_WITH_SERVER,
-				AWAITING_TIMEOUT);
+				context->timeout_secs);
 
 		wsi->mode = LWSCM_HTTP_SERVING;
 
